package org.wordpress.android.ui.reader.adapters;

import android.annotation.SuppressLint;
import android.content.Context;
import android.graphics.Typeface;
import android.graphics.drawable.Drawable;
import android.os.AsyncTask;
import android.text.Spannable;
import android.text.SpannableStringBuilder;
import android.text.style.StyleSpan;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.widget.ImageView;
import android.widget.TextView;

import androidx.annotation.NonNull;
import androidx.recyclerview.widget.RecyclerView;

import org.jetbrains.annotations.NotNull;
import org.wordpress.android.R;
import org.wordpress.android.WordPress;
import org.wordpress.android.analytics.AnalyticsTracker;
import org.wordpress.android.datasets.ReaderPostTable;
import org.wordpress.android.datasets.ReaderTagTable;
import org.wordpress.android.fluxc.store.AccountStore;
import org.wordpress.android.fluxc.store.SiteStore;
import org.wordpress.android.models.ReaderPost;
import org.wordpress.android.models.ReaderPostDiscoverData;
import org.wordpress.android.models.ReaderPostList;
import org.wordpress.android.models.ReaderTag;
import org.wordpress.android.ui.reader.ReaderActivityLauncher;
import org.wordpress.android.ui.reader.ReaderConstants;
import org.wordpress.android.ui.reader.ReaderInterfaces;
import org.wordpress.android.ui.reader.ReaderInterfaces.OnFollowListener;
import org.wordpress.android.ui.reader.ReaderInterfaces.OnPostListItemButtonListener;
import org.wordpress.android.ui.reader.ReaderTypes;
import org.wordpress.android.ui.reader.ReaderTypes.ReaderPostListType;
import org.wordpress.android.ui.reader.actions.ReaderActions;
import org.wordpress.android.ui.reader.actions.ReaderTagActions;
import org.wordpress.android.ui.reader.discover.ReaderCardUiState;
import org.wordpress.android.ui.reader.discover.ReaderCardUiState.ReaderPostUiState;
import org.wordpress.android.ui.reader.discover.ReaderPostCardActionType;
import org.wordpress.android.ui.reader.discover.ReaderPostMoreButtonUiStateBuilder;
import org.wordpress.android.ui.reader.discover.ReaderPostUiStateBuilder;
import org.wordpress.android.ui.reader.discover.viewholders.ReaderPostViewHolder;
import org.wordpress.android.ui.reader.models.ReaderBlogIdPostId;
import org.wordpress.android.ui.reader.tracker.ReaderTab;
import org.wordpress.android.ui.reader.tracker.ReaderTracker;
import org.wordpress.android.ui.reader.utils.ReaderXPostUtils;
import org.wordpress.android.ui.reader.views.ReaderGapMarkerView;
import org.wordpress.android.ui.reader.views.ReaderSiteHeaderView;
import org.wordpress.android.ui.reader.views.ReaderTagHeaderView;
import org.wordpress.android.ui.reader.views.ReaderTagHeaderViewUiState.ReaderTagHeaderUiState;
import org.wordpress.android.ui.reader.views.uistates.FollowButtonUiState;
import org.wordpress.android.ui.utils.UiHelpers;
import org.wordpress.android.util.AppLog;
import org.wordpress.android.util.AppLog.T;
import org.wordpress.android.util.ColorUtils;
import org.wordpress.android.util.extensions.ContextExtensionsKt;
import org.wordpress.android.util.DisplayUtils;
import org.wordpress.android.util.GravatarUtils;
import org.wordpress.android.util.NetworkUtils;
import org.wordpress.android.util.SiteUtils;
import org.wordpress.android.util.ToastUtils;
import org.wordpress.android.util.image.BlavatarShape;
import org.wordpress.android.util.image.ImageManager;
import org.wordpress.android.util.image.ImageType;

import java.util.HashSet;

import javax.inject.Inject;

import kotlin.Unit;
import kotlin.jvm.functions.Function0;
import kotlin.jvm.functions.Function1;
import kotlin.jvm.functions.Function2;
import kotlin.jvm.functions.Function3;

public class ReaderPostAdapter extends RecyclerView.Adapter<RecyclerView.ViewHolder> {
    private final ImageManager mImageManager;
    private final UiHelpers mUiHelpers;
    private ReaderTag mCurrentTag;
    private long mCurrentBlogId;
    private long mCurrentFeedId;
    private int mGapMarkerPosition = -1;

    private final int mPhotonWidth;
    private final int mPhotonHeight;
    private final int mAvatarSzSmall;

    private boolean mCanRequestMorePosts;

    @NonNull private final ReaderTypes.ReaderPostListType mPostListType;
    @NonNull private String mSource;
    private final ReaderPostList mPosts = new ReaderPostList();
    private final HashSet<String> mRenderedIds = new HashSet<>();

    private ReaderInterfaces.OnPostListItemButtonListener mOnPostListItemButtonListener;
    private ReaderInterfaces.OnFollowListener mFollowListener;
    private ReaderInterfaces.OnPostSelectedListener mPostSelectedListener;
    private ReaderInterfaces.DataLoadedListener mDataLoadedListener;
    private ReaderActions.DataRequestedListener mDataRequestedListener;
    private ReaderSiteHeaderView.OnBlogInfoLoadedListener mBlogInfoLoadedListener;

    // the large "tbl_posts.text" column is unused here, so skip it when querying
    private static final boolean EXCLUDE_TEXT_COLUMN = true;
    private static final int MAX_ROWS = ReaderConstants.READER_MAX_POSTS_TO_DISPLAY;

    private static final int VIEW_TYPE_POST = 0;
    private static final int VIEW_TYPE_XPOST = 1;
    private static final int VIEW_TYPE_SITE_HEADER = 2;
    private static final int VIEW_TYPE_TAG_HEADER = 3;
    private static final int VIEW_TYPE_GAP_MARKER = 4;
    private static final int VIEW_TYPE_REMOVED_POST = 5;

    private static final long ITEM_ID_HEADER = -1L;
    private static final long ITEM_ID_GAP_MARKER = -2L;

    private static final float READER_FEATURED_IMAGE_ASPECT_RATIO = 16 / 9f;

    private final boolean mIsMainReader;

    @Inject AccountStore mAccountStore;
    @Inject SiteStore mSiteStore;
    @Inject ReaderPostUiStateBuilder mReaderPostUiStateBuilder;
    @Inject ReaderPostMoreButtonUiStateBuilder mReaderPostMoreButtonUiStateBuilder;
    @Inject ReaderTracker mReaderTracker;

    public String getSource() {
        return mSource;
    }

    /*
     * cross-post
     */
    private class ReaderXPostViewHolder extends RecyclerView.ViewHolder {
        private final ImageView mImgAvatar;
        private final ImageView mImgBlavatar;
        private final TextView mTxtTitle;
        private final TextView mTxtSubtitle;

        ReaderXPostViewHolder(View itemView) {
            super(itemView);
            View postContainer = itemView.findViewById(R.id.post_container);
            mImgAvatar = itemView.findViewById(R.id.image_avatar);
            mImgBlavatar = itemView.findViewById(R.id.image_blavatar);
            mTxtTitle = itemView.findViewById(R.id.text_title);
            mTxtSubtitle = itemView.findViewById(R.id.text_subtitle);

            postContainer.setOnClickListener(v -> {
                int position = getAdapterPosition();
                ReaderPost post = getItem(position);
                if (mPostSelectedListener != null && post != null) {
                    mPostSelectedListener.onPostSelected(post);
                }
            });
        }
    }

    private static class ReaderRemovedPostViewHolder extends RecyclerView.ViewHolder {
        final View mPostContainer;

        private final ViewGroup mRemovedPostContainer;
        private final TextView mTxtRemovedPostTitle;
        private final TextView mUndoRemoveAction;

        ReaderRemovedPostViewHolder(View itemView) {
            super(itemView);
            mPostContainer = itemView.findViewById(R.id.post_container);
            mTxtRemovedPostTitle = itemView.findViewById(R.id.removed_post_title);
            mRemovedPostContainer = itemView.findViewById(R.id.removed_item_container);
            mUndoRemoveAction = itemView.findViewById(R.id.undo_remove);
        }
    }

    private static class SiteHeaderViewHolder extends RecyclerView.ViewHolder {
        private final ReaderSiteHeaderView mSiteHeaderView;

        SiteHeaderViewHolder(View itemView) {
            super(itemView);
            mSiteHeaderView = (ReaderSiteHeaderView) itemView;
        }
    }

    private static class TagHeaderViewHolder extends RecyclerView.ViewHolder {
        private final ReaderTagHeaderView mTagHeaderView;

        TagHeaderViewHolder(View itemView) {
            super(itemView);
            mTagHeaderView = (ReaderTagHeaderView) itemView;
        }

        public void onBind(ReaderTagHeaderUiState uiState) {
            mTagHeaderView.updateUi(uiState);
        }
    }

    private static class GapMarkerViewHolder extends RecyclerView.ViewHolder {
        private final ReaderGapMarkerView mGapMarkerView;

        GapMarkerViewHolder(View itemView) {
            super(itemView);
            mGapMarkerView = (ReaderGapMarkerView) itemView;
        }
    }

    @Override
    public int getItemViewType(int position) {
        if (position == 0 && hasSiteHeader()) {
            // first item is a ReaderSiteHeaderView
            return VIEW_TYPE_SITE_HEADER;
        } else if (position == 0 && hasTagHeader()) {
            // first item is a ReaderTagHeaderView
            return VIEW_TYPE_TAG_HEADER;
        } else if (position == mGapMarkerPosition) {
            return VIEW_TYPE_GAP_MARKER;
        } else {
            ReaderPost post = getItem(position);
            if (post != null && post.isXpost()) {
                return VIEW_TYPE_XPOST;
            } else if (post != null && isBookmarksList() && !post.isBookmarked) {
                return VIEW_TYPE_REMOVED_POST;
            } else {
                return VIEW_TYPE_POST;
            }
        }
    }

    @Override
    public @NonNull RecyclerView.ViewHolder onCreateViewHolder(@NonNull ViewGroup parent, int viewType) {
        Context context = parent.getContext();
        View postView;
        switch (viewType) {
            case VIEW_TYPE_SITE_HEADER:
                ReaderSiteHeaderView readerSiteHeaderView = new ReaderSiteHeaderView(context);
                readerSiteHeaderView.setOnFollowListener(mFollowListener);
                return new SiteHeaderViewHolder(readerSiteHeaderView);

            case VIEW_TYPE_TAG_HEADER:
                return new TagHeaderViewHolder(new ReaderTagHeaderView(context));

            case VIEW_TYPE_GAP_MARKER:
                return new GapMarkerViewHolder(new ReaderGapMarkerView(context));

            case VIEW_TYPE_XPOST:
                postView = LayoutInflater.from(context).inflate(R.layout.reader_cardview_xpost, parent, false);
                return new ReaderXPostViewHolder(postView);
            case VIEW_TYPE_REMOVED_POST:
                postView = LayoutInflater.from(context).inflate(R.layout.reader_cardview_removed_post, parent, false);
                return new ReaderRemovedPostViewHolder(postView);
            default:
                return new ReaderPostViewHolder(mUiHelpers, mImageManager, mReaderTracker, parent);
        }
    }

    @Override
    public void onBindViewHolder(@NonNull RecyclerView.ViewHolder holder, int position) {
        if (holder instanceof ReaderPostViewHolder) {
            renderPost(position, (ReaderPostViewHolder) holder, false);
        } else if (holder instanceof ReaderXPostViewHolder) {
            renderXPost(position, (ReaderXPostViewHolder) holder);
        } else if (holder instanceof ReaderRemovedPostViewHolder) {
            renderRemovedPost(position, (ReaderRemovedPostViewHolder) holder);
        } else if (holder instanceof SiteHeaderViewHolder) {
            SiteHeaderViewHolder siteHolder = (SiteHeaderViewHolder) holder;
            siteHolder.mSiteHeaderView.setOnBlogInfoLoadedListener(mBlogInfoLoadedListener);
            if (isDiscover()) {
                siteHolder.mSiteHeaderView.loadBlogInfo(
                        ReaderConstants.DISCOVER_SITE_ID,
                        0,
                        mSource
                );
            } else {
                siteHolder.mSiteHeaderView.loadBlogInfo(
                        mCurrentBlogId,
                        mCurrentFeedId,
                        mSource
                );
            }
        } else if (holder instanceof TagHeaderViewHolder) {
            TagHeaderViewHolder tagHolder = (TagHeaderViewHolder) holder;
            renderTagHeader(mCurrentTag, tagHolder, true);
        } else if (holder instanceof GapMarkerViewHolder) {
            GapMarkerViewHolder gapHolder = (GapMarkerViewHolder) holder;
            gapHolder.mGapMarkerView.setCurrentTag(mCurrentTag);
        }
    }

    private void renderTagHeader(
            ReaderTag currentTag,
            TagHeaderViewHolder tagHolder,
            Boolean isFollowButtonEnabled
    ) {
        if (currentTag == null) {
            return;
        }
        Function0<Unit> onFollowButtonClicked = () -> {
            toggleFollowButton(tagHolder.itemView.getContext(), currentTag, tagHolder);
            return Unit.INSTANCE;
        };

        ReaderTagHeaderUiState uiState = new ReaderTagHeaderUiState(
                currentTag.getLabel(),
                new FollowButtonUiState(
                        onFollowButtonClicked,
                        ReaderTagTable.isFollowedTagName(currentTag.getTagSlug()),
                        isFollowButtonEnabled,
                        true
                )
        );
        tagHolder.onBind(uiState);
    }

    private void toggleFollowButton(
            Context context,
            @NotNull ReaderTag currentTag,
            TagHeaderViewHolder tagHolder
    ) {
        if (!NetworkUtils.checkConnection(context)) {
            return;
        }

        final boolean isAskingToFollow = !ReaderTagTable.isFollowedTagName(currentTag.getTagSlug());

        final String slugForTracking = currentTag.getTagSlug();

        ReaderActions.ActionListener listener = succeeded -> {
            if (!succeeded) {
                int errResId = isAskingToFollow ? R.string.reader_toast_err_add_tag
                        : R.string.reader_toast_err_remove_tag;
                ToastUtils.showToast(context, errResId);
            } else {
                if (isAskingToFollow) {
                    mReaderTracker.trackTag(
                            AnalyticsTracker.Stat.READER_TAG_FOLLOWED,
                            slugForTracking,
                            mSource
                    );
                } else {
                    mReaderTracker.trackTag(
                            AnalyticsTracker.Stat.READER_TAG_UNFOLLOWED,
                            slugForTracking,
                            mSource
                    );
                }
            }
            renderTagHeader(currentTag, tagHolder, true);
        };

        boolean success;
        boolean isLoggedIn = mAccountStore.hasAccessToken();
        if (isAskingToFollow) {
            success = ReaderTagActions.addTag(mCurrentTag, listener, isLoggedIn);
        } else {
            success = ReaderTagActions.deleteTag(mCurrentTag, listener, isLoggedIn);
        }

        if (isLoggedIn && success) {
            renderTagHeader(currentTag, tagHolder, false);
        }
    }

    private void renderXPost(int position, ReaderXPostViewHolder holder) {
        final ReaderPost post = getItem(position);
        if (post == null) {
            return;
        }

        mImageManager
                .loadIntoCircle(holder.mImgAvatar, ImageType.AVATAR,
                        GravatarUtils.fixGravatarUrl(post.getPostAvatar(), mAvatarSzSmall));

        mImageManager.loadIntoCircle(holder.mImgBlavatar,
                SiteUtils.getSiteImageType(post.isP2orA8C(), BlavatarShape.CIRCULAR),
                GravatarUtils.fixGravatarUrl(post.getBlogImageUrl(), mAvatarSzSmall));

        holder.mTxtTitle.setText(ReaderXPostUtils.getXPostTitle(post));
        holder.mTxtSubtitle.setText(ReaderXPostUtils.getXPostSubtitleHtml(post));

        checkLoadMore(position);
    }

    private void renderRemovedPost(final int position, final ReaderRemovedPostViewHolder holder) {
        final ReaderPost post = getItem(position);
        final Context context = holder.mRemovedPostContainer.getContext();
        holder.mTxtRemovedPostTitle.setText(createTextForRemovedPostContainer(post, context));
        Drawable drawable =
                ColorUtils.applyTintToDrawable(context, R.drawable.ic_undo_white_24dp,
                        ContextExtensionsKt.getColorResIdFromAttribute(context, R.attr.colorPrimary));
        holder.mUndoRemoveAction.setCompoundDrawablesWithIntrinsicBounds(drawable, null, null, null);
        holder.mPostContainer.setOnClickListener(v -> undoPostUnbookmarked(post));
    }

    private void undoPostUnbookmarked(final ReaderPost post) {
        if (!post.isBookmarked) {
            mOnPostListItemButtonListener.onButtonClicked(post, ReaderPostCardActionType.BOOKMARK);
        }
    }

    private void renderPost(final int position, final ReaderPostViewHolder holder, boolean showMoreMenu) {
        final ReaderPost post = getItem(position);
        ReaderPostListType postListType = getPostListType();
        if (post == null) {
            return;
        }
        Context ctx = holder.getViewContext();
        Function3<Long, Long, ReaderPostCardActionType, Unit> onButtonClicked =
                (postId, blogId, type) -> {
                    mOnPostListItemButtonListener.onButtonClicked(post, type);
                    renderPost(position, holder, false);
                    return Unit.INSTANCE;
                };
        Function2<Long, Long, Unit> onItemClicked = (postId, blogId) -> {
            if (mPostSelectedListener != null) {
                mPostSelectedListener.onPostSelected(post);
            }
            return Unit.INSTANCE;
        };
        Function1<ReaderCardUiState, Unit> onItemRendered = (item) -> {
            checkLoadMore(position);

            // if we haven't already rendered this post and it has a "railcar" attached to it, add it
            // to the rendered list and record the TrainTracks render event
            if (post.hasRailcar() && !mRenderedIds.contains(post.getPseudoId())) {
                mRenderedIds.add(post.getPseudoId());
                mReaderTracker.trackRailcar(post.getRailcarJson());
            }
            return Unit.INSTANCE;
        };
        Function2<Long, Long, Unit> onDiscoverSectionClicked = (postId, blogId) -> {
            ReaderPostDiscoverData discoverData = post.getDiscoverData();
            switch (discoverData.getDiscoverType()) {
                case EDITOR_PICK:
                    if (mPostSelectedListener != null) {
                        mPostSelectedListener.onPostSelected(post);
                    }
                    break;
                case SITE_PICK:
                    if (discoverData.getBlogId() != 0) {
                        ReaderActivityLauncher.showReaderBlogPreview(
                                ctx,
                                discoverData.getBlogId(),
                                post.isFollowedByCurrentUser,
                                mSource,
                                mReaderTracker
                        );
                    } else if (discoverData.hasBlogUrl()) {
                        ReaderActivityLauncher.openUrl(ctx, discoverData.getBlogUrl());
                    }
                    break;
                case OTHER:
                    // noop
                    break;
            }
            return Unit.INSTANCE;
        };
        Function1<ReaderPostUiState, Unit> onMoreButtonClicked = (uiState) -> {
            renderPost(position, holder, true);
            return Unit.INSTANCE;
        };

        Function1<ReaderPostUiState, Unit> onMoreDismissed = (uiState) -> {
            renderPost(position, holder, false);
            return Unit.INSTANCE;
        };

        Function2<Long, Long, Unit> onVideoOverlayClicked = (postId, blogId) -> {
            ReaderActivityLauncher.showReaderVideoViewer(ctx, post.getFeaturedVideo());
            return Unit.INSTANCE;
        };

        Function2<Long, Long, Unit> onPostHeaderClicked = (postId, blogId) -> {
            ReaderActivityLauncher.showReaderBlogPreview(
                    ctx,
                    post,
                    mSource,
                    mReaderTracker
            );
            return Unit.INSTANCE;
        };

        Function1<String, Unit> onTagItemClicked = (tagSlug) -> {
            // noop
            return Unit.INSTANCE;
        };

        ReaderPostUiState uiState = mReaderPostUiStateBuilder
                .mapPostToUiStateBlocking(
                        mSource,
                        post,
                        false,
                        mPhotonWidth,
                        mPhotonHeight,
                        postListType,
                        onButtonClicked,
                        onItemClicked,
                        onItemRendered,
                        onDiscoverSectionClicked,
                        onMoreButtonClicked,
                        onMoreDismissed,
                        onVideoOverlayClicked,
                        onPostHeaderClicked,
                        onTagItemClicked,
                        showMoreMenu ? mReaderPostMoreButtonUiStateBuilder
                                .buildMoreMenuItemsBlocking(post, onButtonClicked) : null
                );
        holder.onBind(uiState);
    }

    /*
     * if we're nearing the end of the posts, fire request to load more
     */
    private void checkLoadMore(int position) {
        if (mCanRequestMorePosts
            && mDataRequestedListener != null
            && (position >= getItemCount() - 1)) {
            mDataRequestedListener.onRequestData();
        }
    }

    // ********************************************************************************************

    public ReaderPostAdapter(
            Context context,
            ReaderPostListType postListType,
            ImageManager imageManager,
            UiHelpers uiHelpers,
            boolean isMainReader
    ) {
        super();
        ((WordPress) context.getApplicationContext()).component().inject(this);
        this.mImageManager = imageManager;
        mPostListType = postListType;
        mSource = mReaderTracker.getSource(mPostListType);
        mUiHelpers = uiHelpers;
        mAvatarSzSmall = context.getResources().getDimensionPixelSize(R.dimen.avatar_sz_small);
        mIsMainReader = isMainReader;

        int displayWidth = DisplayUtils.getWindowPixelWidth(context);
        int cardMargin = context.getResources().getDimensionPixelSize(R.dimen.reader_card_margin);
        mPhotonWidth = displayWidth - (cardMargin * 2);
        mPhotonHeight = (int) (mPhotonWidth / READER_FEATURED_IMAGE_ASPECT_RATIO);

        setHasStableIds(true);
    }

    private boolean hasHeader() {
        return hasSiteHeader() || hasTagHeader();
    }

    private boolean hasSiteHeader() {
        return !mIsMainReader && (isDiscover() || getPostListType() == ReaderTypes.ReaderPostListType.BLOG_PREVIEW);
    }

    private boolean hasTagHeader() {
        return (getPostListType() == ReaderPostListType.TAG_PREVIEW) && !isEmpty();
    }

    private boolean isDiscover() {
        return mCurrentTag != null && mCurrentTag.isDiscover();
    }

    public void setOnPostListItemButtonListener(OnPostListItemButtonListener listener) {
        mOnPostListItemButtonListener = listener;
    }

    public void setOnFollowListener(OnFollowListener listener) {
        mFollowListener = listener;
    }

    public void setOnPostSelectedListener(ReaderInterfaces.OnPostSelectedListener listener) {
        mPostSelectedListener = listener;
    }

    public void setOnDataLoadedListener(ReaderInterfaces.DataLoadedListener listener) {
        mDataLoadedListener = listener;
    }

    public void setOnDataRequestedListener(ReaderActions.DataRequestedListener listener) {
        mDataRequestedListener = listener;
    }

    public void setOnBlogInfoLoadedListener(ReaderSiteHeaderView.OnBlogInfoLoadedListener listener) {
        mBlogInfoLoadedListener = listener;
    }

    private ReaderTypes.ReaderPostListType getPostListType() {
        return (mPostListType != null ? mPostListType : ReaderTypes.DEFAULT_POST_LIST_TYPE);
    }

    // used when the viewing tagged posts
    public void setCurrentTag(ReaderTag tag) {
        mSource = mReaderTracker.getSource(mPostListType, ReaderTab.transformTagToTab(tag));
        if (!ReaderTag.isSameTag(tag, mCurrentTag)) {
            mCurrentTag = tag;
            mRenderedIds.clear();
            reload();
        }
    }

    public boolean isCurrentTag(ReaderTag tag) {
        return ReaderTag.isSameTag(tag, mCurrentTag);
    }

    // used when the list type is ReaderPostListType.BLOG_PREVIEW
    public void setCurrentBlogAndFeed(long blogId, long feedId) {
        if (blogId != mCurrentBlogId || feedId != mCurrentFeedId) {
            mCurrentBlogId = blogId;
            mCurrentFeedId = feedId;
            mRenderedIds.clear();
            reload();
        }
    }

    public void clear() {
        mGapMarkerPosition = -1;
        if (!mPosts.isEmpty()) {
            mPosts.clear();
            notifyDataSetChanged();
        }
    }

    public void refresh() {
        loadPosts();
    }

    /*
     * same as refresh() above but first clears the existing posts
     */
    public void reload() {
        clear();
        loadPosts();
    }

    @SuppressWarnings("deprecation")
    private void loadPosts() {
        if (mIsTaskRunning) {
            AppLog.w(AppLog.T.READER, "reader posts task already running");
            return;
        }
        new LoadPostsTask().executeOnExecutor(AsyncTask.THREAD_POOL_EXECUTOR);
    }

    private ReaderPost getItem(int position) {
        if (position == getHeaderPosition() && hasHeader()) {
            return null;
        }
        if (position == mGapMarkerPosition) {
            return null;
        }

        int arrayPos = position - getItemPositionOffset();

        if (mGapMarkerPosition > -1 && position > mGapMarkerPosition) {
            arrayPos--;
        }

        if (mPosts.size() <= arrayPos) {
            AppLog.d(T.READER, "Trying to read an element out of bounds of the posts list");
            return null;
        }

        return mPosts.get(arrayPos);
    }

    private int getItemPositionOffset() {
        return hasHeader() ? 1 : 0;
    }

    private int getHeaderPosition() {
        return hasHeader() ? 0 : -1;
    }

    @Override
    public int getItemCount() {
        int size = mPosts.size();
        if (mGapMarkerPosition != -1) {
            size++;
        }
        if (hasHeader()) {
            size++;
        }
        return size;
    }

    public boolean isEmpty() {
        return (mPosts == null || mPosts.size() == 0);
    }

    private boolean isBookmarksList() {
        return (getPostListType() == ReaderPostListType.TAG_FOLLOWED
                && (mCurrentTag != null && mCurrentTag.isBookmarked()));
    }

    @Override
    public long getItemId(int position) {
        switch (getItemViewType(position)) {
            case VIEW_TYPE_TAG_HEADER:
            case VIEW_TYPE_SITE_HEADER:
                return ITEM_ID_HEADER;
            case VIEW_TYPE_GAP_MARKER:
                return ITEM_ID_GAP_MARKER;
            default:
                ReaderPost post = getItem(position);
                return post != null ? post.getStableId() : 0;
        }
    }

    /**
     * Creates 'Removed [post title]' text, with the '[post title]' in bold.
     */
    @NonNull
    private SpannableStringBuilder createTextForRemovedPostContainer(ReaderPost post, Context context) {
        String removedString = context.getString(R.string.removed);
        String removedPostTitle = removedString + " " + post.getTitle();
        SpannableStringBuilder str = new SpannableStringBuilder(removedPostTitle);
        str.setSpan(new StyleSpan(Typeface.BOLD), removedString.length(), removedPostTitle.length(),
                Spannable.SPAN_EXCLUSIVE_EXCLUSIVE);
        return str;
    }

    public void setFollowStatusForBlog(long blogId, boolean isFollowing) {
        ReaderPost post;
        for (int i = 0; i < mPosts.size(); i++) {
            post = mPosts.get(i);
            if (post.blogId == blogId && post.isFollowedByCurrentUser != isFollowing) {
                post.isFollowedByCurrentUser = isFollowing;
                mPosts.set(i, post);
                notifyItemChanged(i);
            }
        }
    }

    public void removeGapMarker() {
        if (mGapMarkerPosition == -1) {
            return;
        }

        int position = mGapMarkerPosition;
        mGapMarkerPosition = -1;
        if (position < getItemCount()) {
            notifyItemRemoved(position);
        }
    }

    /*
     * AsyncTask to load posts in the current tag
     */
    private boolean mIsTaskRunning = false;

    @SuppressWarnings("deprecation")
    @SuppressLint("StaticFieldLeak")
    private class LoadPostsTask extends AsyncTask<Void, Void, Boolean> {
        private ReaderPostList mAllPosts;

        private boolean mCanRequestMorePostsTemp;
        private int mGapMarkerPositionTemp;

        @Override
        protected void onPreExecute() {
            mIsTaskRunning = true;
        }

        @Override
        protected void onCancelled() {
            mIsTaskRunning = false;
        }

        @Override
        protected Boolean doInBackground(Void... params) {
            int numExisting;
            switch (getPostListType()) {
                case TAG_PREVIEW:
                case TAG_FOLLOWED:
                case SEARCH_RESULTS:
                    mAllPosts = ReaderPostTable.getPostsWithTag(mCurrentTag, MAX_ROWS, EXCLUDE_TEXT_COLUMN);
                    numExisting = ReaderPostTable.getNumPostsWithTag(mCurrentTag);
                    break;
                case BLOG_PREVIEW:
                    if (mCurrentFeedId != 0) {
                        mAllPosts = ReaderPostTable.getPostsInFeed(mCurrentFeedId, MAX_ROWS, EXCLUDE_TEXT_COLUMN);
                        numExisting = ReaderPostTable.getNumPostsInFeed(mCurrentFeedId);
                    } else {
                        mAllPosts = ReaderPostTable.getPostsInBlog(mCurrentBlogId, MAX_ROWS, EXCLUDE_TEXT_COLUMN);
                        numExisting = ReaderPostTable.getNumPostsInBlog(mCurrentBlogId);
                    }
                    break;
                default:
                    return false;
            }

            if (mPosts.isSameListWithBookmark(mAllPosts)) {
                return false;
            }

            // if we're not already displaying the max # posts, enable requesting more when
            // the user scrolls to the end of the list
            mCanRequestMorePostsTemp = (numExisting < ReaderConstants.READER_MAX_POSTS_TO_DISPLAY);

            // determine whether a gap marker exists - only applies to tagged posts
            mGapMarkerPositionTemp = getGapMarkerPosition();

            return true;
        }

        private int getGapMarkerPosition() {
            if (!getPostListType().isTagType()) {
                return -1;
            }

            ReaderBlogIdPostId gapMarkerIds = ReaderPostTable.getGapMarkerIdsForTag(mCurrentTag);
            if (gapMarkerIds == null) {
                return -1;
            }

            int gapMarkerPostPosition = mAllPosts.indexOfIds(gapMarkerIds);
            int gapMarkerPosition = -1;
            if (gapMarkerPostPosition > -1) {
                // remove the gap marker if it's on the last post (edge case but
                // it can happen following a purge)
                if (gapMarkerPostPosition == mAllPosts.size() - 1) {
                    AppLog.w(AppLog.T.READER, "gap marker at/after last post, removed");
                    ReaderPostTable.removeGapMarkerForTag(mCurrentTag);
                } else {
                    // we want the gap marker to appear *below* this post
                    gapMarkerPosition = gapMarkerPostPosition + 1;
                    // increment it if there are custom items at the top of the list (header)
                    gapMarkerPosition += getItemPositionOffset();
                    AppLog.d(AppLog.T.READER, "gap marker at position " + gapMarkerPostPosition);
                }
            }
            return gapMarkerPosition;
        }

        @Override
        protected void onPostExecute(Boolean result) {
            if (result) {
                ReaderPostAdapter.this.mGapMarkerPosition = mGapMarkerPositionTemp;
                ReaderPostAdapter.this.mCanRequestMorePosts = mCanRequestMorePostsTemp;
                mPosts.clear();
                mPosts.addAll(mAllPosts);
                notifyDataSetChanged();
            }

            if (mDataLoadedListener != null) {
                mDataLoadedListener.onDataLoaded(isEmpty());
            }

            mIsTaskRunning = false;
        }
    }
}
